# coding=utf-8

import os, sys
from distutils.version import LooseVersion, StrictVersion
import numpy as np
sys.path.append('../')

import logging
logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger('GalPaK_Tests:')


## GENERAL PACKAGES
import unittest
import matplotlib
matplotlib.use('Agg')

import galpak
from galpak import GalPaK3D, DefaultModel
from galpak import GalaxyParameters, GalaxyParametersError
from galpak import MUSE
from galpak.instruments import *
from galpak import GaussianPointSpreadFunction, MoffatPointSpreadFunction
from galpak import HyperspectralCube as Cube
try:
    import corner
    corner_true = True
except:
    corner_true = False

from tests.tools.numpy_ndarray_assertions import NumpyNdArrayAssertions

#Python3 compatibility
try:
  basestring
except NameError:
  basestring = str


class Something(object):
    """ Mock class for testing invalid input """
    pass


class GalPaK3DTest(unittest.TestCase, NumpyNdArrayAssertions):
    """
    Some tests plot images, and they WILL PAUSE THE RUNNER until you close the images. (CTRL+W = Close)
    """

    longMessage = True

    root_folder = os.path.abspath(os.path.join(
        os.path.dirname(os.path.abspath(__file__)), os.pardir, os.pardir
    ))

    fits_folder = os.path.join(root_folder, 'data/input')
    fits_mpdaf_filename = os.path.join(fits_folder, 'cube/subcube_mpdaf.fits')
    fits_muse_filename = os.path.join(fits_folder, 'cube/subcube_MUSE.fits')
    fits_real_muse_filename = os.path.join(fits_folder,
                                           'cube/zoom_oii_subcube_3Dcontinuum_subtracted.fits')
    fits_real_muse_variance = os.path.join(fits_folder, 'zoomVariance_oii_subcube_3Dcontinuum_subtracted.fits')

    def assertEqualish(self, expected, actual, expected_similarity=99.):
        """
        Assert that actual is similar to expected, at expected_similarity (in %).
        This assertion can be improved :
        - Failure message
        - Input type-check
        """
        actual_precision = 100. * (1 - np.abs(expected - actual) / expected)
        self.assertTrue(actual_precision >= expected_similarity)

    def test_init_with_not_a_cube_nor_a_filename(self):
        with self.assertRaises(TypeError):
            g = GalPaK3D(Something())

    def test_init_with_empty_cube(self):
        with self.assertRaises(ValueError):
            g = GalPaK3D(Cube())

    def test_init_with_unfit_cube(self):
        with self.assertRaises(Exception):
            g = GalPaK3D(Cube.from_file('this_is_not_a_fits_file'))

    def test_init_with_unfit_filename(self):
        with self.assertRaises(Exception):
            g = GalPaK3D('this_is_not_a_fits_file')

    def test_init_with_mpdaf_cube(self):
        logger.info('Test init with mpdaf')
        try:
            import mpdaf
            cube = mpdaf.obj.Cube(filename=self.fits_muse_filename)
            g = GalPaK3D(cube)  # should not raise
            self.assertIsNotNone(g.variance_cube)
        except ImportError:
            pass

    def test_init_with_real_muse_cube(self):
        logger.info("Test init with muse cube")
        g = GalPaK3D(self.fits_real_muse_filename,
                     variance=self.fits_real_muse_variance,
                     seeing=0.65)
        g.run_mcmc(max_iterations=100, verbose=False)

    def test_init_with_all_of_our_input_fits(self):
        for filename in os.listdir(self.fits_folder):
            if os.path.isfile(filename):
                filename = os.path.join(self.fits_folder, filename)
                g = GalPaK3D(filename)  # should not raise; if it does, use sanitizer

    def test_init_with_default_context(self):
        logger.info("test_default_context")
        g = GalPaK3D(self.fits_muse_filename)
        self.assertIsInstance(g.instrument, MUSE,
                              "Default instrument should be MUSE")

    def test_init_with_default_model(self):
        logger.info("test_init_default_model")

        g = GalPaK3D(self.fits_muse_filename)
        m = DefaultModel(redshift=0.8)
        g.run_mcmc(max_iterations=10, model=m, verbose=False)

    def test_reference(self):
        logger.info("test against reference")
        #def _test_model(self, galaxy=None, instrument=None, model=None, ref_clean=None, ref_conv=None):
        #"""
        #test model calculation against a reference cube
        #:param galaxy:
        #:param instrument: Instrument MUSE(psf_fwhm=0.8) [default]
        #:param model: DefaultModel [default]
        #:param ref_cube:
        #:return:
        #"""

        fits_folder = os.path.join(self.root_folder, 'data/input/ref')
        ref_clean =  os.path.join(fits_folder, 'ref_cube_clean_gpk188.fits')
        ref_conv  = os.path.join(fits_folder, 'ref_cube_conv_gpk188.fits')


        instrument = MUSE(psf_fwhm=0.7, psf_ba=1.0, psf_pa=0)
        galaxy = None

        if isinstance(ref_clean, basestring):
            ref_clean = Cube.from_file(ref_clean)
        elif isinstance(ref_clean, Cube):
            pass
        else:
            raise NameError("ref_clean should be a string or a HyperCube")

        gk1 = GalPaK3D(ref_clean, instrument=instrument)

        if isinstance(ref_conv, basestring):
            ref_conv = Cube.from_file(ref_conv)
        elif isinstance(ref_conv, Cube):
            pass

        gk2 = GalPaK3D(ref_conv, instrument=instrument)

        if LooseVersion(galpak.__version__) >= LooseVersion('1.9'):
            model = DefaultModel(flux_profile='exponential', rotation_curve='arctan', \
                                 thickness_profile='gaussian', dispersion_profile='thick')
            gk1.model=model
            gk2.model=model
        else:
            model={}
            model['flux_profile']='exponential'
            model['rotation_curve']='arctan'
            model['disk_dispersion']='thick'
            model['hz_profile']='gaussian'

            gk1.disk_model=model
            gk2.disk_model=model



        if  galaxy is None:
            galaxy = GalaxyParameters(
                x=15, y=15, z=15,
                flux = 1,
                radius = 2.5,
                inclination = 45,
                pa = 280,
                turnover_radius=0.5,
                maximum_velocity=180,
                velocity_dispersion=40
            )


        clean_cube = gk1.create_clean_cube(galaxy, ref_clean.data.shape)
        convolved_cube = gk2.create_convolved_cube(galaxy, ref_conv.data.shape)

        diff_clean = np.nansum( (clean_cube - ref_clean).data)
        diff_conv = np.nansum( (convolved_cube - ref_conv).data)

        test = (diff_clean) !=0 or (diff_conv) !=0
        if test:
            logger.warning("Difference found with reference cube!!! %.3e" %  (diff_clean))
            logger.warning("Difference found with ref convolved cube!!! %.3e" % (diff_conv))
        else:
            logger.info("Test passed successfully !! No differences found")

        self.assertTrue(test, "Model still works")


    def test_run_mcmc_output(self):
        logger.info("test run mcmc")
        glpk3d = GalPaK3D(self.fits_muse_filename)
        galaxy = glpk3d.run_mcmc(max_iterations=10, verbose=False)

        self.assertIsInstance(galaxy, GalaxyParameters,
                              "Deconvolution should return galaxy parameters")
        self.assertIsInstance(galaxy.stdev, GalaxyParametersError,
                              "Deconvolution should return galaxy parameters with stdev")
        self.assertIsNotNone(glpk3d.deconvolved_cube,
                             "Deconvolution should fill deconvolved_cube property")
        self.assertIsNotNone(glpk3d.convolved_cube,
                             "Deconvolution should fill convolved_cube property")
        self.assertIsNotNone(glpk3d.residuals_cube,
                             "Deconvolution should fill residuals_cube property")
        self.assertIsNotNone(glpk3d.chain,
                             "Deconvolution should fill chain property")
        self.assertIsNotNone(glpk3d.psf3d,
                             "Deconvolution should fill psf3d property")
        self.assertIsNotNone(glpk3d.acceptance_rate,
                             "Deconvolution should fill acceptance_rate property")
        self.assertIsNotNone(glpk3d.galaxy,
                             "Deconvolution should fill galaxy property")
        self.assertIsNotNone(glpk3d.stdev,
                             "Deconvolution should fill stdev property")
        self.assertIs(glpk3d.galaxy, galaxy,
                      "Returned galaxy should be a view to galaxy property")
        self.assertIs(glpk3d.stdev, galaxy.stdev,
                      "Returned galaxy's stdev should be a view to stdev property")

    def test_run_two_instances(self):
        logger.info("test with 2 instances")
        glpk3d_a = GalPaK3D(self.fits_muse_filename)
        galaxy_a = glpk3d_a.run_mcmc(max_iterations=10, verbose=False)
        glpk3d_b = GalPaK3D(self.fits_muse_filename)
        galaxy_b = glpk3d_b.run_mcmc(max_iterations=10, verbose=False)

        self.assertIsInstance(glpk3d_a.galaxy, GalaxyParameters)
        self.assertIsInstance(glpk3d_b.galaxy, GalaxyParameters)

    def test_run_with_variance(self):
        logger.info("test with variance")
        g = GalPaK3D(self.fits_real_muse_filename, variance=self.fits_real_muse_variance, seeing=0.65)
        p = g.run_mcmc(max_iterations=10, verbose=False)

        self.assertIsInstance(p,GalaxyParameters)

    def test_run_with_boundaries(self):

        glpk3d = GalPaK3D(self.fits_muse_filename)
        glpk3d.instrument.psf.fwhm = 0.6
        min_boundaries = GalaxyParameters(x=14.)
        max_boundaries = GalaxyParameters(x=16.)
        galaxy = glpk3d.run_mcmc(max_iterations=10,
                                 min_boundaries=min_boundaries,
                                 max_boundaries=max_boundaries,
                                 verbose=False)

        self.assertTrue(galaxy.x >= min_boundaries.x,
                        "Galaxy's x should be more than min_boundaries's x")
        self.assertTrue(galaxy.x <= max_boundaries.x,
                        "Galaxy's x should be less than max_boundaries's x")

    def test_run_with_doublet(self):
        g = GalPaK3D(self.fits_real_muse_filename)
        model = DefaultModel(line=galpak.OII)
        p = g.run_mcmc(max_iterations=30, verbose=False, model=model)

        self.assertIsInstance(p,GalaxyParameters)

    def test_Cosmology(self):

        m = DefaultModel(redshift=1)
        self.assertIsInstance(m.DeltaVir,float)

    def test_Extention_DefaultModel(self):

        for rc in DefaultModel.CURVE_VALID:
            m = DefaultModel(rotation_curve=rc, redshift=1)
            g = m.Parameters()
            g.from_ndarray(np.arange(len(g)))
            m.set_v22(g)
            m.set_dvdx(g)


    def test_instruments(self):

        muse_nfm = MUSENFM()
        self.assertEqual(muse_nfm.cube_default_xy_step, 0.025)
        muse_wfm = MUSEWFM()
        self.assertEqual(muse_wfm.cube_default_xy_step, 0.2)
        muse = MUSE()
        self.assertEqual(muse.cube_default_xy_step, muse_wfm.cube_default_xy_step,
                         "Default MUSE Field Mode is Wide")

        # ... many additional tests can go here
        kmos = KMOS()
        self.assertEqual(kmos.cube_default_xy_step, 0.2)

        alma = ALMA()
        self.assertEqual(alma.cube_default_xy_step, None)

        alma = ALMA(pixscale=0.1)
        self.assertEqual(alma.cube_default_xy_step, 0.1)

        sinfok1 = SINFOK250()
        self.assertEqual(sinfok1.cube_default_xy_step, 0.125)

        sinfok2 = SINFOK100()
        self.assertEqual(sinfok2.cube_default_xy_step, 0.05)

        sinfoj1 = SINFOJ250()
        self.assertEqual(sinfoj1.cube_default_xy_step, 0.125)

        sinfoj2 = SINFOJ100()
        self.assertEqual(sinfoj2.cube_default_xy_step, 0.05)

        osiris = OSIRIS()
        self.assertEqual(osiris.cube_default_xy_step, 0.035)

        harmoni = HARMONI(pixscale=0.030, psf_fwhm=1.0, lsf_fwhm=0.3)
        self.assertEqual(harmoni.cube_default_xy_step, 0.030)



    def test_instruments_psf(self):
        glpk3d = GalPaK3D(self.fits_muse_filename)
        self.assertIsInstance(glpk3d.instrument.psf, GaussianPointSpreadFunction,
                              "Gaussian PSF is the default for MUSE when unspecified")

        #for inst in [MUSEWFM(), MUSENFM(), KMOS(), SINFOK250(), ALMA(pixscale=0.1), OSIRIS(), HARMONI(pixscale=30,lsf_fwhm=0.3)]:
        inst = MUSEWFM()
        glpk3d = GalPaK3D(self.fits_muse_filename, instrument=inst)
        glpk3d._save_to_file('test_instrument.txt', glpk3d.instrument.__str__(), True)
        self.assertTrue(os.path.isfile('test_instrument.txt'))
        new = GalPaK3D(glpk3d.cube, instrument='test_instrument.txt')
        self.assertEqual(new.instrument.__str__(),glpk3d.instrument.__str__())
        os.remove('test_instrument.txt')

        inst = MUSENFM()
        glpk3d = GalPaK3D(Cube(data=np.ones((20,30,30))), instrument=inst)
        glpk3d._save_to_file('test_instrument.txt', glpk3d.instrument.__str__(), True)
        self.assertTrue(os.path.isfile('test_instrument.txt'))
        new = GalPaK3D(glpk3d.cube, instrument='test_instrument.txt')
        self.assertEqual(new.instrument.__str__(), glpk3d.instrument.__str__())
        os.remove('test_instrument.txt')

        inst = ALMA(pixscale=0.1)
        glpk3d = GalPaK3D(Cube(data=np.ones((20,30,30))), instrument=inst, crval3=1.0, cdelt3=1.0)
        glpk3d._save_to_file('test_instrument.txt', glpk3d.instrument.__str__(), True)
        self.assertTrue(os.path.isfile('test_instrument.txt'))
        new = GalPaK3D(Cube(data=np.ones((20,30,30))), instrument='test_instrument.txt', crval3=1.0, cdelt3=1.0)
        self.assertEqual(new.instrument.__str__(), glpk3d.instrument.__str__())
        os.remove('test_instrument.txt')

        inst = KMOS()
        glpk3d = GalPaK3D(Cube(data=np.ones((20,30,30))), instrument=inst, crval3=1.0)
        glpk3d._save_to_file('test_instrument.txt', glpk3d.instrument.__str__(), True)
        self.assertTrue(os.path.isfile('test_instrument.txt'))
        new = GalPaK3D(Cube(data=np.ones((20,30,30))), instrument='test_instrument.txt', crval3=1.0)
        self.assertEqual(new.instrument.__str__(), glpk3d.instrument.__str__())
        os.remove('test_instrument.txt')

        inst = OSIRIS()
        glpk3d = GalPaK3D(Cube(data=np.ones((20,30,30))), instrument=inst)
        glpk3d._save_to_file('test_instrument.txt', glpk3d.instrument.__str__(), True)
        self.assertTrue(os.path.isfile('test_instrument.txt'))
        new = GalPaK3D(glpk3d.cube, instrument='test_instrument.txt')
        self.assertEqual(new.instrument.__str__(), glpk3d.instrument.__str__())
        os.remove('test_instrument.txt')

        inst = HARMONI(pixscale=0.03, psf_fwhm=0.2, lsf_fwhm=1.2)
        glpk3d = GalPaK3D(Cube(data=np.ones((20,30,30))), instrument=inst, crval3=1.0, cdelt3=0.0002)
        glpk3d._save_to_file('test_instrument.txt', glpk3d.instrument.__str__(), True)
        self.assertTrue(os.path.isfile('test_instrument.txt'))
        new = GalPaK3D(Cube(data=np.ones((20,30,30))), instrument='test_instrument.txt', crval3=1.0, cdelt3=0.0002)
        self.assertEqual(new.instrument.__str__(), glpk3d.instrument.__str__())
        os.remove('test_instrument.txt')

        #inst = Generic(psf_fwhm=1, psf_pa=0, psf_ba=1, lsf_fwhm=1.2)
        #glpk3d = GalPaK3D(Cube(data=np.ones((20,30,30))), instrument=inst)
        #glpk3d._save_to_file('test_instrument.txt', glpk3d.instrument.__str__(), True)
        #self.assertTrue(os.path.isfile('test_instrument.txt'))
        #new = GalPaK3D(glpk3d.cube, instrument='test_instrument.txt')
        #self.assertEqual(new.instrument.__str__(), glpk3d.instrument.__str__())
        #os.remove('test_instrument.txt')

        # glpk3d = GalPaK3D(self.fits_muse_filename, instrument=MUSE(psf=None))
        # self.assertIsNone(glpk3d.instrument.psf,
        #                   "PSF should be de-activated when psf=None is specified")

        glpk3d = GalPaK3D(self.fits_muse_filename, instrument=MUSE(psf=MoffatPointSpreadFunction(fwhm=1,beta=2.5,ba=1,pa=0)))
        self.assertIsInstance(glpk3d.instrument.psf, MoffatPointSpreadFunction,
                              "Use specified PSF")

    def test_wavelength_conversion(self):
        glpk3d = GalPaK3D(self.fits_muse_filename)
        params = glpk3d.run_mcmc(max_iterations=100, verbose=False)

        expected = 6743.46774459
        actual = glpk3d.cube.wavelength_of(params.z)
        self.assertEqualish(expected, actual, 95.)

    def test_api(self):

        g = galpak.run(self.fits_real_muse_filename, max_iterations=250, verbose=False)

        self.assertIsInstance(g, GalPaK3D)

    def test_api_auto(self):

        g = galpak.autorun(self.fits_real_muse_filename, max_iterations=250, verbose=False)

        self.assertIsInstance(g, GalPaK3D)

    def test_plotting(self):
        logger.info("test plotting")
        glpk3d = GalPaK3D(self.fits_real_muse_filename)
        glpk3d.run_mcmc(max_iterations=300, verbose=False)

        # MCMC
        glpk3d.plot_mcmc('test_plot_mcmc.png')
        self.assertTrue(os.path.isfile('test_plot_mcmc.png'))
        os.remove('test_plot_mcmc.png')
        #glpk3d.plot_mcmc()
        #glpk3d.plot_mcmc(sort_by_chi=True)

        # IMAGES
        glpk3d.plot_images('test_plot_images.png')
        self.assertTrue(os.path.isfile('test_plot_images.png'))
        os.remove('test_plot_images.png')

        # CoRNER
        glpk3d.plot_corner('test_plot_corner.png')
        if corner_true:
            self.assertTrue(os.path.isfile('test_plot_corner.png'))
            os.remove('test_plot_corner.png')

        # Geweke
        glpk3d.plot_geweke('test_plot_geweke.png')
        self.assertTrue(os.path.isfile('test_plot_geweke.png'))
        os.remove('test_plot_geweke.png')

        #ModelUtilities
        glpk3d.model.plot_vprofile(glpk3d.galaxy, filename='test_rotcurve.png')
        self.assertTrue(os.path.isfile('test_rotcurve.png'))
        os.remove('test_rotcurve.png')

        glpk3d.model.plot_SBprofile(glpk3d.galaxy, filename='test_SBprofile.png')
        self.assertTrue(os.path.isfile('test_SBprofile.png'))
        os.remove('test_SBprofile.png')
        # ANIMATION
        #glpk3d.make_animation()

    def test_save(self):
        logger.info("test save")
        os.system('rm -f test_*')

        glpk3d = GalPaK3D(self.fits_real_muse_filename)
        glpk3d.run_mcmc(max_iterations=300, verbose=False)
        glpk3d.save('test', overwrite=True)

        self.assertTrue(os.path.isfile('test_galaxy_parameters.txt'))
        os.remove('test_galaxy_parameters.txt')
        self.assertTrue(os.path.isfile('test_galaxy_parameters.dat'))
        os.remove('test_galaxy_parameters.dat')
        self.assertTrue(os.path.isfile('test_chain.dat'))
        os.remove('test_chain.dat')
        self.assertTrue(os.path.isfile('test_convolved_cube.fits'))
        os.remove('test_convolved_cube.fits')
        self.assertTrue(os.path.isfile('test_deconvolved_cube.fits'))
        os.remove('test_deconvolved_cube.fits')
        self.assertTrue(os.path.isfile('test_residuals_cube.fits'))
        os.remove('test_residuals_cube.fits')
        self.assertTrue(os.path.isfile('test_3Dkernel.fits'))
        os.remove('test_3Dkernel.fits')

        self.assertTrue(os.path.isfile('test_obs_disp_map.fits'))
        os.remove('test_obs_disp_map.fits')
        self.assertTrue(os.path.isfile('test_obs_flux_map.fits'))
        os.remove('test_obs_flux_map.fits')
        self.assertTrue(os.path.isfile('test_obs_vel_map.fits'))
        os.remove('test_obs_vel_map.fits')

        self.assertTrue(os.path.isfile('test_true_disp_map.fits'))
        os.remove('test_true_disp_map.fits')
        self.assertTrue(os.path.isfile('test_true_flux_map.fits'))
        os.remove('test_true_flux_map.fits')
        self.assertTrue(os.path.isfile('test_true_vel_map.fits'))
        os.remove('test_true_vel_map.fits')

        if glpk3d.error_maps:
            self.assertTrue(os.path.isfile('test_true_disp_map_error.fits'))
            os.remove('test_true_disp_map_error.fits')
            self.assertTrue(os.path.isfile('test_true_flux_map_error.fits'))
            os.remove('test_true_flux_map_error.fits')
            self.assertTrue(os.path.isfile('test_true_vel_map_error.fits'))
            os.remove('test_true_vel_map_error.fits')

        self.assertTrue(os.path.isfile('test_mcmc.png'))
        self.assertTrue(os.path.isfile('test_mcmc.pdf'))
        os.remove('test_mcmc.png')
        os.remove('test_mcmc.pdf')


        self.assertTrue(os.path.isfile('test_images.png'))
        self.assertTrue(os.path.isfile('test_images.pdf'))
        os.remove('test_images.png')
        os.remove('test_images.pdf')

        self.assertTrue(os.path.isfile('test_true_maps.png'))
        self.assertTrue(os.path.isfile('test_true_maps.pdf'))
        os.remove('test_true_maps.png')
        os.remove('test_true_maps.pdf')

        self.assertTrue(os.path.isfile('test_obs_maps.png'))
        self.assertTrue(os.path.isfile('test_obs_maps.pdf'))
        os.remove('test_obs_maps.png')
        os.remove('test_obs_maps.pdf')

        self.assertTrue(os.path.isfile('test_true_Vrot.dat'))
        os.remove('test_true_Vrot.dat')

        self.assertTrue(os.path.isfile('test_instrument.txt'))
        os.remove('test_instrument.txt')
        self.assertTrue(os.path.isfile('test_run_parameters.txt'))
        os.remove('test_run_parameters.txt')

        if corner_true:
            self.assertTrue(os.path.isfile('test_corner.pdf'))
            os.remove('test_corner.pdf')
            self.assertTrue(os.path.isfile('test_corner.png'))
            os.remove('test_corner.png')

        self.assertTrue(os.path.isfile('test_geweke.pdf'))
        os.remove('test_geweke.pdf')
        self.assertTrue(os.path.isfile('test_geweke.png'))
        os.remove('test_geweke.png')

        self.assertTrue(os.path.isfile('test_galaxy_parameters_convergence.dat'))
        os.remove('test_galaxy_parameters_convergence.dat')

        self.assertTrue(os.path.isfile('test_model.txt'))
        os.remove('test_model.txt')

        self.assertTrue(os.path.isfile('test_stats.dat'))
        os.remove('test_stats.dat')

        self.assertTrue(os.path.isfile('test_rotcurve.pdf'))
        os.remove('test_rotcurve.pdf')
        self.assertTrue(os.path.isfile('test_rotcurve.png'))
        os.remove('test_rotcurve.png')

    # def test_load_from_chain_dat(self):
    #     glpk3d = GalPaK3D(self.fits_real_muse_filename)
    #     glpk3d.import_chain('test_chain.dat')
    #     glpk3d.plot_mcmc()






import benchmark
import numpy as np

#Benchmark Report
#================
#
#BenchmarkPowerOfTwo
#-------------------
#
#                   name | rank |  runs |      mean |        sd | timesBaseline
#------------------------|------|-------|-----------|-----------|--------------
#       double asterisks |    1 | 1e+05 | 3.889e-05 |  4.66e-06 |           1.0
#explicit multiplication |    2 | 1e+05 | 4.024e-05 | 9.693e-06 | 1.03474171064
#            numpy power |    3 | 1e+05 | 0.0004267 | 2.959e-05 | 10.9717454015
#
#Each of the above 300000 runs were run in random, non-consecutive order by
#`benchmark` v0.1.5 (http://jspi.es/benchmark) with Python 2.7.5+
#Linux-3.11.0-12-generic-x86_64 on 2014-02-13 13:26:33.


class BenchmarkPowerOfTwo(benchmark.Benchmark):

    each = 10000  # configure number of runs

    def setUp(self):
        # Can also specify tearDown, eachSetUp, and eachTearDown
        self.a = np.random.rand(32, 32, 32)

    def test_explicit_multiplication(self):
        b = self.a * self.a

    def test_double_asterisks(self):
        b = self.a ** 2

    def test_numpy_power(self):
        b = np.power(self.a, 2)


if __name__ == '__main__':
    benchmark.main(format="markdown", numberFormat="%.4g")




import numpy


class NumpyNdArrayAssertions():
    """
    This class is a mixin following the UnitTest naming conventions.
    It is meant to be used along with unittest.TestCase like so :
    class MyTestCase(unittest.TestCase, NumpyNdArrayAssertions):
        ...
    It needs python >= 2.6 and numpy >= 1.6
    """

    def assertArrayEqual(self, expected_array, actual_array, msg=None):
        """
        Fail if actual array is not the same as expected array,
        both element-wise and shape-wise, and don't choke on NaNs.
        """
        standardMsg = "%s\nis not expected\n%s" % (str(actual_array), str(expected_array))
        numpy.testing.assert_array_equal(actual_array, expected_array,
                                         err_msg=self._formatMessage(msg, standardMsg),
                                         verbose=self.longMessage)

    def assertArrayEquals(self, expected_array, actual_array, msg=None):
        """
        Alias for assertArrayEqual
        """
        self.assertArrayEqual(expected_array, actual_array, msg)

    def assertArrayEqualish(self, expected_array, actual_array, expected_similarity=99., msg=None):
        """
        Fails if average similarity is less than expected_similarity (in %)
        """
        a = numpy.nan_to_num(actual_array)
        e = numpy.nan_to_num(expected_array)

        actual_precision = 100. * (1 - numpy.nan_to_num(numpy.abs(e - a) / e))
        actual_precision = numpy.sum(actual_precision) / actual_precision.size

        standardMsg = "%f %% is less than %f %%" % (actual_precision, expected_similarity)

        self.assertTrue(actual_precision >= expected_similarity, self._formatMessage(msg, standardMsg))

        return actual_precision


# This file's purpose is to make MPDAF's distutils `setup.py` work.

from .__version__ import __version__

from .galpak3d import *
from .galaxy_parameter import *
from .galaxy_parameters import *
from .model_sersic3d import ModelSersic
DefaultModel = ModelSersic
from .instruments import *
from .spread_functions import *
from .hyperspectral_cube import *
from .string_stdout import *
from .api import *
from .ansi_colors import *
from .math_utils import  *



# coding=utf-8

import os
from distutils.version import LooseVersion, StrictVersion
import benchmark

import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('GalPaK_bench:')


## GENERAL PACKAGES

from galpak import __version__
from galpak import GalPaK3D, DiskModel
from galpak import GalaxyParameters
from galpak import MUSE
from galpak import HyperspectralCube as Cube

#Python3 compatibility
try:
  basestring
except NameError:
  basestring = str

class Benchmark_galpak(benchmark.Benchmark):

    each = 100  # configure number of runs

    root_folder = os.path.abspath(os.path.join(
        os.path.dirname(os.path.abspath(__file__)), os.pardir, os.pardir
    ))


    fits_folder = os.path.join(root_folder, 'data/input/ref')
    ref_clean =  os.path.join(fits_folder, 'ref_cube_clean_gpk188.fits')
    ref_conv  = os.path.join(fits_folder, 'ref_cube_conv_gpk188.fits')


    instrument = MUSE(psf_fwhm=0.7, psf_ba=1.0, psf_pa=0)
    galaxy = None

    if isinstance(ref_clean, basestring):
        ref_clean = Cube.from_file(ref_clean)
    elif isinstance(ref_clean, Cube):
        pass
    else:
        raise NameError("ref_clean should be a string or a HyperCube")

    gk1 = GalPaK3D(ref_clean, instrument=instrument)

    if isinstance(ref_conv, basestring):
        ref_conv = Cube.from_file(ref_conv)
    elif isinstance(ref_conv, Cube):
        pass

    gk2 = GalPaK3D(ref_conv, instrument=instrument)

    if StrictVersion(__version__) >= StrictVersion('1.9'):
        model = DiskModel()
        gk1.model=model
        gk2.model=model
    else:
        model={}
        model['flux_profile']='exponential'
        model['rotation_curve']='arctan'
        model['disk_dispersion']='thick'
        model['hz_profile']='gaussian'

        gk1.disk_model=model
        gk2.disk_model=model
    #verbose
    gk1.verbose=False
    gk2.verbose=False


    if  galaxy is None:
        galaxy = GalaxyParameters(
            x=15, y=15, z=15,
            flux = 1,
            radius = 2.5,
            inclination = 45,
            pa = 280,
            turnover_radius=0.5,
            maximum_velocity=180,
            velocity_dispersion=40
        )

    def test_reference(self):
                #def _test_model(self, galaxy=None, instrument=None, model=None, ref_clean=None, ref_conv=None):
                #"""
                #test model calculation against a reference cube
                #:param galaxy:
                #:param instrument: Instrument MUSE(psf_fwhm=0.8) [default]
                #:param model: DiskModel [default]
                #:param ref_cube:
                #:return:
                #"""

        self.gk1.create_clean_cube(self.galaxy, self.ref_clean.data.shape)
        self.gk2.create_convolved_cube(self.galaxy, self.ref_conv.data.shape)

        #diff_clean = np.nansum( (clean_cube - ref_clean).data)
        #diff_conv = np.nansum( (convolved_cube - ref_conv).data)

        #test = (diff_clean) !=0 or (diff_conv) !=0
        #if test:
        #    logger.warning("Difference found with reference cube!!! %.3e" %  (diff_clean))
        #    logger.warning("Difference found with ref convolved cube!!! %.3e" % (diff_conv))
        #else:
        #    logger.info("Test passed successfully !! No differences found")

    def test_runner(self):

        self.gk1.run_mcmc(max_iterations=100,verbose=False,save_error_maps=False)

if __name__ == '__main__':
    benchmark.main(format="markdown", numberFormat="%.4g")




# -*- coding: utf-8 -*-

import numpy as np
from scipy import special
from astropy.io import fits
from copy import deepcopy
try:
    import bottleneck as bn
except ImportError:
    import numpy as bn

from .galaxy_parameters import  GalaxyParameters
from .model_utilities import modelPlots, ModelExt
from .model_class import Model
from .math_utils import flux_sersic

# LOGGING CONFIGURATION
import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('GalPaK: DiskModel ')

class ModelSersic(Model,modelPlots, ModelExt):
    """
    The first and default galpak `Model` (when unspecified).
    It simulates a simple disk galaxy using `DiskParameters`.

    flux_profile: 'exponential' | 'gaussian' | 'de_vaucouleurs' | 'sersicN2' | 'user'
            The flux profile of the observed galaxy. The default is 'exponential'.
            See http://en.wikipedia.org/wiki/Sersic%27s_law
            If 'user' provided flux map, specify flux_image. Currently unsupported

    thickness_profile:  'gaussian' [default] | 'exponential' | 'sech2' | 'none'
        The perpendicular flux profile of the disk galaxy.
        The default is 'gaussian'.
        If none. the galaxy will become a cylinder.

    rotation_curve: 'arctan'  |  'exponential' | 'tanh' [default] |
                 'isothermal' | 'NFW' | 'Burkert' |
                 'mass' | 'Freeman' | 'Spano'
        The profile of the velocity v(r) can be in (with X=r/rV where rV=turnover radius):
            - arctan : Vmax arctan(X),
            - tanh : Vmax tanh(X),
            - exponential : inverted exponential, Vmax (1-exp(-X))
            - isothermal : for an isothermal halo, Vmax [1-arctan(X)/X)]^0.5
            - NFW :  for a NFW halo, Vmax {[ln(1+X) - X/(1+X)]/X}^0.5
            - Burkert : for a Burkert halo, Vmax {[ln(1 + X) + 0.5 ln (1 + X^2) - arctan(X)]/X}^0.5
            - Freeman : a Freeman disk with Vmax=V2.2
            - mass : a constant light-to-mass ratio v(r)=sqrt(G m(<r) / r); where m(<r) is the cumulative mass profile with m(<r) = \int I(r) 2 pi r dr
            - Spano: for a Spano 2008 profile, Vmax {arcsinh(X)/X - 1/sqrt(1+X2)}
                    See https://doi.org/10.1093/mnras/stz1969

        The default is 'tanh'.

    dispersion_profile: 'thick' [default] | 'thin'
        The local disk dispersion from the rotation curve and disk thickness
        from Binney & Tremaine 2008, see Genzel et al. 2008.
        GalPak has 3 components for the dispersion:
            - a component from the rotation curve arising from mixing velocities
              of a disk with non-zero thickness.
            - a component from the local disk dispersion specified by
              `disk_dispersion`
            - a spatially constant dispersion,
              which is the output parameter `velocity_dispersion`.

    line: None[default] | dict to fit doublets, use a dictionary with
            - line['wave']=[l, lref] eg. [3726.2, 3728.9] # Observed or Rest
            - line['ratio']=[0.8, 1.0] eg. [0.8, 1.0] # The primary line for redshifts is the reddest

    redshift: float
            The redshift of the observed galaxy, used in mass calculus.
            Will override the redshift provided at init ; this is a convenience parameter.
            If this is not set anywhere, GalPak will not try to compute the mass.

    """

    # fixme: use integers instead of strings here and enjoy the performance gain

    FLUX_VALID = ['gaussian', 'exponential', 'de_vaucouleurs', 'sersicN2', 'user']

    THICKNESS_VALID = [ 'gaussian', 'exponential', 'sech2', 'none']

    DISPERSION_VALID = [ 'thick', 'thin']

    CURVE_VALID = ['arctan', 'exponential', 'tanh', \
                   'isothermal', 'NFW', 'Burkert', 'Freeman', 'mass', 'Spano' ]

    logger = logger

    def __init__(self,
                 flux_profile='exponential',
                 thickness_profile='gaussian',
                 dispersion_profile='thick',
                 rotation_curve= 'tanh',
                 flux_image=None,
                 line=None,
                 aspect=None,
                 redshift=None,
                 pixscale=None,
                 cosmology='planck15'):
        if flux_profile in self.FLUX_VALID:
            self.flux_profile = flux_profile
        else:
            raise ValueError("Flux parameter should be one of self.FLUX_VALID",self.FLUX_VALID)

        if rotation_curve in self.CURVE_VALID:
            self.rotation_curve = rotation_curve
        else:
            raise ValueError("Rotation curve should be one of self.CURVE_VALID", self.CURVE_VALID)

        if thickness_profile in self.THICKNESS_VALID:
            self.thickness_profile = thickness_profile
        else:
            raise ValueError("Thickness profile should be one of self.THICKNESS_VALID", self.THICKNESS_VALID)

        if dispersion_profile in self.DISPERSION_VALID:
            self.dispersion_profile = dispersion_profile
        else:
            raise ValueError('Dispersion profile should be one of self.DISPERSION_VALID', self.DISPERSION_VALID)


        self.line = line
        self.aspect = aspect
        self.redshift = redshift
        self.pixscale = pixscale
        self.cosmology= cosmology
        self.logger = logging.getLogger('GalPaK: DiskModel')

        if self.rotation_curve == 'Freeman' and self.flux_profile != 'exponential':
            self.logger.warning("You have selected a Freeman disk and a flux profile that is not exponential. Use this at your own risks.")

        if self.rotation_curve == 'mass':
            self.logger.warning("You have selected the mass disk profile; which only works for gaussian/exponential/sersicN2 flux profile")

        #fluxmap user
        if self.flux_profile == 'user':
            if isinstance(flux_image, str):
                try:
                    self.flux_map_user = fits.open(flux_image)['DATA'].data
                except KeyError:
                    self.flux_map_user = fits.open(flux_image)[0].data
                else:
                    raise ValueError('Flux user image should have data in DATA or extention 0')
            elif isinstance(flux_image, np.ndarray):
                self.flux_map_user = flux_image
            else:
                self.logger.error("Flux image should be an array or a string")

        if self.redshift != None :
            if self.cosmology is not None:
                self.set_cosmology(self.cosmology)
            else:
                self.set_cosmology() #using default

        if self.aspect is not None:
            self.q = self.aspect
        else:
            self.q = 0.15 #default aspect ratio

        #self.logger.info("Set up model: %s" % (self.__str__()))


    def initial_parameters(self, runner):
        """
        Returns an instance of a class extending `ModelParameters`.
        You may omit parameters, they will be automatically set to the mean of
        the max and min.
        """
        # Default initial parameters
        init = self.parameters_class()(
            radius=3.0,
            turnover_radius=1.0,
            inclination=30,
            pa=np.random.uniform(0,360),
            maximum_velocity=100.0,
            velocity_dispersion=5.0,
            flux=runner.flux_est
        )

        return init

    def min_boundaries(self, runner):
        """
        Returns an instance of a class extending `ModelParameters`.
        You MUST provide all the parameters.
        """
                # Default boundaries
        return GalaxyParameters(
            x=runner.cube.shape[2] / 3.,
            y=runner.cube.shape[1] / 3.,
            z=runner.cube.shape[0] / 3.,
            flux=0.,
            radius=0.8,
            inclination=-5,
            pa=-40,
            turnover_radius=0.01,
            maximum_velocity=10.,
            velocity_dispersion=0.
        )


    def max_boundaries(self, runner):
        """
        Returns an instance of a class extending `ModelParameters`.
        You MUST provide all the parameters.
        """
        # changed max radius to 3/8th of cube size.
        return GalaxyParameters(
            x=runner.cube.shape[2] * 2 / 3.,
            y=runner.cube.shape[1] * 2 / 3.,
            z=runner.cube.shape[0] * 2 / 3.,
            flux=3. * runner.flux_est,
            radius= 0.5 * (runner.cube.shape[1]+runner.cube.shape[2]) * 3 / 8.,
            inclination=90.5,
            pa=400,
            turnover_radius=0.5 * (runner.cube.shape[1]+runner.cube.shape[2]) * 3 / 8. / 2.,
            maximum_velocity=350.,
            velocity_dispersion=180.
        )


    def setup_random_amplitude(self, amplitude):
        """
        Mutates the random `amplitude` of the parameter jump with custom logic.
        This is called during the setup of the MCMC runner.
        The amplitude is already filled with :
            sqrt((min_boundaries - max_boundaries) ** 2 / 12.) * p / v
        where
            p = number of parameters in the model
            v = number of voxels in the cube
        """
        amplitude['flux'] *= 0.5  # flux since we start at ~flux since flux is an integrated quantity
        amplitude['pa'] *= 1.5
        amplitude['maximum_velocity'] *= 1.5 #
        amplitude['velocity_dispersion'] *= 1.5  # velocity dispersion
        amplitude *= 3       # fixme: document this coefficient

    #CUSTOM METHODS
    def set_flux_profile(self, galaxy, radius_cube):
        """
        creates a flux_cube in 3D (x,y,z)
        returns flux_cube
        """

        rhalf = galaxy['radius']

        if self.flux_profile == 'de_vaucouleurs':
            flux_cube = flux_sersic(radius_cube, rhalf, 4.0)
        elif self.flux_profile == 'gaussian':
            flux_cube = flux_sersic(radius_cube, rhalf, 0.5)
        elif self.flux_profile == 'exponential':
            flux_cube = flux_sersic(radius_cube, rhalf, 1.0)
        elif self.flux_profile == 'sersicN2':
            flux_cube = flux_sersic(radius_cube, rhalf, 2.0)
        elif self.flux_profile == 'user':
            if len(radius_cube.shape)==1:
                #then measure 1D profile from parameters
                flux_cube = self._user_profile(parameters, radius_cube)
            elif len(radius_cube.shape)==3:
                flux_cube = self.flux_map_user[None, :, :]  # @fixme need thickness?
            else:
                raise ValueError("set profile with a 1d vector or 3d cube")
        else:
            raise ValueError("Flux profile is invalid. Should be one of '%s' ." % self.FLUX_VALID)

        return flux_cube


    def set_velocity_profile(self, galaxy, radius_cube):
        """
        creates a velocity profile for circular orbits
        return v_profile
        """
        max_vel = galaxy['maximum_velocity']
        rt = galaxy['turnover_radius']

        x = radius_cube / rt

        # Flat rotation curve
        if self.rotation_curve == 'arctan':
            v_profile = np.arctan(x) * max_vel * 2. / np.pi
        #
        elif self.rotation_curve == 'tanh':
            v_profile = max_vel * np.tanh(x)
        # For Feng's inv_exp profile
        elif self.rotation_curve == 'exponential':
            v_profile = max_vel * (1.0 - np.exp(-x))
        # ISOTHERMAL rotation curve;
        elif self.rotation_curve == 'isothermal' :
            v_profile = max_vel * np.sqrt(1 - np.arctan(x)/x)
        elif self.rotation_curve == 'NFW':
            v_sq =  ( np.log(1+x) - x / (1+x) ) / x #peaks at x=2.16
            v_sq /= ( np.log(1+2.16) - 2.16 / (1+2.16)) / 2.16
            v_profile = max_vel * np.sqrt(v_sq)
        elif self.rotation_curve == 'Burkert':
            v_sq =  (np.log(1+x)+0.5*np.log(1+x**2)-np.arctan(x)) / x # peaks at x=3.24
            v_sq /= (np.log(4.24)+0.5*np.log(1+3.24**2)-np.arctan(3.24)) / 3.24
            v_profile = max_vel * np.sqrt(v_sq)
        elif self.rotation_curve == 'mass':
            v_profile = self._fv_newton(radius_cube, galaxy['radius'], max_vel)
        elif self.rotation_curve == 'Freeman':
            #Using Freeman disk. Vmax = V2.2
            v_sq = self._v_disk_sq(galaxy, radius_cube)
            v_profile = np.sqrt(v_sq)
        elif self.rotation_curve == 'Spano':
            # From Spano et al. 2008: https://doi.org/10.1111/j.1365-2966.2007.12545.x
            # and Hernandez-Hernandez et al. 2019 https://doi.org/10.1093/mnras/stz1969
            v_sq = 1/x * np.vectorize(np.math.asinh)(x) - 1. / (1+x**2)
            v_profile = max_vel * np.sqrt(v_sq)
        else:
            raise NotImplementedError("Rotational Curve not supported. Should be '%s' " %
                                      self.CURVE_VALID)

        return v_profile

    #Private methods

    def _v_disk_sq(self, parameters, radius):
        ##disk Freeman model
        y = 1.68 * radius / parameters['radius'] / 2.
        B = special.i0(y) * special.k0(y) \
            - special.i1(y) * special.k1(y) #
        # Swinbank
        #  V(r)^2 = 0.5 (GMd/Rd) * (3.6*x)^2 * (I0K0-I1K1)
        #
        # Eq. 2.165. BinneyTremaine 2008
        # Vsq = 2 * (G Md/Rd) y**2 [I0K0-I1K1]
        # Vsq = 4 * (G Md/Rd)  0.5 * y**2 [I0K0-I1K1]

        B1 = 0.1934  #y^2 B(y) peaks at y=1.1;
        v_disk_sq =  parameters['maximum_velocity']**2. * (y)**2. * B / B1
        # 2 Vdisk_sq * B1 = Vmax_sq

        return v_disk_sq



from galpak import GalPaK3D, MUSEWFM
from os import pardir
from os.path import abspath, dirname, join

# filename = 'contsubCube_Hb_MUSE_J223256.08-603414.17_z0.56'
filename = 'contsubCube_O3_MUSE_J223256.08-603414.17_z0.56'

root_folder = abspath(join(dirname(abspath(__file__)), pardir))
fits_folder = join(root_folder, 'data/input')
fits_file = join(fits_folder, filename + '.fits')

glpk3d = GalPaK3D(fits_file, seeing=0.7, instrument=MUSEWFM(lsf_fwhm=2.519/1e4))
glpk3d.run_mcmc(max_iterations=10000, verbose=True)

# ANIMATION
glpk3d.film_images(filename, frames_skipped=5)

glpk3d.save(filename, overwrite=True)


# -*- coding: utf-8 -*-

import numpy as np
from matplotlib import pyplot as plt

# LOGGING CONFIGURATION
from .galpak3d import *
from .galaxy_parameters import GalaxyParameters

import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('GalPaK: DiskUtilities:')


#################################################################
#
class ModelExt:
    ##Methods to update galaxy parameters

    def set_v22(self, galaxy):
        """
        Computes velocity at 2.2 Re
        :param galaxy: GalaxyParameters | Table
        :return: return V(2.2 Re)

        This Model method will add/update the galaxy property
        """
        #V at 2.2 x Rd
        x = 2.2 * galaxy['radius'] / 1.68
        _v22 = self.set_velocity_profile(galaxy, x)
        #set v22 for galaxy
        if isinstance(galaxy, GalaxyParameters):
            logger.info("Adding v22 property to GalaxyParameters class")
            galaxy.set_v22(_v22)
        elif isinstance(galaxy, Table):
            logger.info("Adding v22 property to Table class")
            if 'v22' not in galaxy.colnames:
                galaxy.add_column(_v22,name='v22')
            else:
                galaxy['v22']=_v22
        else:
            raise Exception("set_v22 must use a GalaxyParameters or a Table")

    def set_dvdx(self, galaxy):
        """
        Computes Velocity inner slope at R=0
        :param galaxy: GalaxyParameters | Table
        :return: dv/dx(r=0) using the units of GalaxyParameters (km/s / pix)

        This Model method will add/update the galaxy property
        """
        dx = 0.1
        _dvdx = (self.set_velocity_profile(galaxy, dx)) / (dx)
        if isinstance(galaxy, GalaxyParameters):
            logger.info("Adding dvdx property to GalaxyParameters class")
            galaxy.set_dvdx(_dvdx)
        elif isinstance(galaxy, Table):
            logger.info("Adding dvdx property to Table class")
            if 'dvdx' not in galaxy.colnames:
                galaxy.add_column(_dvdx,name='dvdx')
            else:
                galaxy['dvdx']=_dvdx
        else:
            raise Exception("set_dvdx(p) must use a GalaxyParameters or Table")


#############
### Utilities
##
#@fixme: make tests
class modelPlots:
    pixscale = None

    def plot_velocity(self, parameters, chain=None, figure=3, xmax=None, v_ymax=None, sigma=None,
                      filename=None, legendOff=False, normalizeX=False):
        """
        plot velocity profile utility
            uses _plot_vprofile private method from original Model class

        :param parameters: galaxy GalaxyParameters
        :param pixscale: float | None [default]
        :param normalizeX: boolean [default=False] to normalize x-axis to Rhalf
        :return:
        """

        fig=plt.figure(figure)
        fig.set_figheight(6)
        fig.set_figwidth(9)

        if filename is not None:
            plt.clf()

        plt.subplot(1,2,2)
        self.plot_vprofile(parameters, chain=chain, normalizeX=normalizeX, xmax=xmax, ymax=v_ymax,
                           legendOff=legendOff, figure=None)

        plt.subplot(1,2,1)
        self.plot_SBprofile(parameters, chain=chain, normalizeX=normalizeX, xmax=xmax,
                            legendOff=legendOff, figure=None, sigma=sigma)


        if filename is None:
            plt.show()
        else:
            plt.savefig(filename)
            plt.close()


    def plot_SBprofile(self, galaxy, chain=None, sigma=None, fmt=None, ymax=None, xmax=None, normalizeX=False, legendOff=False, figure=3, filename=None):
        """

        :param galaxy:
        :param model:
        :param normalizeX:
        :return:
        """
        percentile = galaxy.ICpercentile

        if figure is not None:
            plt.figure(figure,figsize=(6,6))
            if filename is not None:
                    plt.clf()

        if self.pixscale is not None:
            pixscale = self.pixscale
        redshift = self.redshift

        if self.redshift is not None:
            kpc = self.kpc
        else:
            logger.warning("redshift not set")

        xpix = np.r_[0.1:galaxy['radius']*5:0.1]

        if self.redshift is not None and self.pixscale is not None:
            x = xpix * kpc * pixscale
            rhalf = galaxy['radius'] * kpc * pixscale
        else:
            x = xpix.copy()
            rhalf = galaxy['radius']

        if normalizeX:
            x = x / rhalf
            rhalf = 1.

        if fmt is None:
            fmt = 'k-'

        f_x = self.set_flux_profile(galaxy, xpix)
        f_x = f_x/np.max(f_x)
        plt.plot(x,f_x,fmt,label='SB(r) ' + self.flux_profile)

        #add errorbar
        if sigma is not None:
            idx = np.arange(0,len(xpix),len(xpix)/20.,dtype=np.int64)
            plt.errorbar(x[idx],f_x[idx],yerr=sigma)

        if chain is not None and self.flux_profile != 'user':
            if not isinstance(chain, Table):
                raise ValueError("Chain must be passed as an astropy Table")
            else:
                med, low, up = self.predict_SBprofile(xpix, chain=chain, percentile=percentile)
                plt.plot(x, up, 'r', lw=1); plt.plot(x, low, 'r', lw=1);
                plt.fill_between(x.flatten(), low, up, color="grey", alpha=0.5, label='%d%%' % (percentile))

        plt.axvline(2*rhalf,ls=':',label= '2 Rhalf')
        sini = np.sin(np.radians(galaxy['inclination']))
        plt.ylabel('SB profile')

        if normalizeX == False:
            if self.redshift is not None and self.pixscale is not None:
                plt.xlabel('x (kpc)')
            else:
                plt.xlabel('x (pix)')
        else:
            plt.xlabel('r/Rhalf')
        plt.xlim(-0.1*rhalf,5*rhalf)
        plt.ylim([0.001,2])
        plt.yscale('log')

        if legendOff is False:
            plt.legend()

        if filename is not None:
            plt.savefig(filename)
            plt.close()

    def plot_vprofile(self, galaxy, chain=None,  normalizeX=False, legendOff=False, ymax=None, xmax=None, fmt=None, figure=3, filename=None):
        """

        :param parameters:
        :param chain:
        :param xmax: None [Default]
        :param normalizeX: False [Default] | True
        :return:
        """
        if figure is not None:
            plt.figure(figure,figsize=(6,6))
            if filename is not None:
                plt.clf()

        percentile = galaxy.ICpercentile

        if self.pixscale is not None:
            pixscale = self.pixscale
        redshift = self.redshift

        if self.redshift is not None:
            kpc = self.kpc

        if xmax is None:
            xpix = np.r_[0.01:galaxy['radius']*5:0.1]
        else:
            xpix = np.r_[0.01:xmax:1.]

        if self.redshift is not None and self.pixscale is not None:
            x = xpix * kpc * pixscale
            rhalf = galaxy['radius'] * kpc * pixscale
            #plt.title('Rhalf %.2f (kpc)' % (galaxy.radius * kpc * pixscale))
        else:
            x = xpix.copy()
            rhalf = galaxy['radius']
            #plt.title('Rhalf %.2f (pix)' % (galaxy.radius))

        if normalizeX:
            x = x / rhalf
            rhalf = 1.

        #plot error if sub_chain
        if chain is not None:
            if not isinstance(chain, Table):
                raise ValueError("Chain must be passed as an astropy Table")
            else:
                mu, low, up = self.predict_velocity(xpix, chain=chain, percentile=percentile)
                plt.plot(x, up, 'r', lw=1); plt.plot(x, low, 'r', lw=1);
                plt.fill_between(x.flatten(), low, up, color="grey", alpha=0.5, label='%d%%' % (percentile))
                fmt='k'
        elif fmt is None:
            fmt='k+'

        v_x = self.set_velocity_profile(galaxy, xpix)
        plt.plot(x,v_x,fmt,label='V(r) ' + self.rotation_curve)

        plt.axvline(2*rhalf,ls=':',label= '2 Rhalf')
        sini = np.sin(np.radians(galaxy['inclination']))
        plt.ylabel('V(km/s) deproj.')

        if self.pixscale is not None and self.redshift is not None:
            Mdyn = self.compute_Mdyn_at_Rhalf(galaxy)
            title = "Mdyn(Re) = %.2f " %(np.log10(Mdyn))
            Mdyn2 = self.compute_Mdyn_at_Rhalf(galaxy,radius=galaxy['radius']*2)
            title += " Mdyn(2Re) = %.2f " %(np.log10(Mdyn2))
        else:
            title = 'Rotation curve'

        ax=plt.gca()
        #plt.text(0.2,0.8,"log Mdyn(R1/2) = %.2f" %(np.log10(Mdyn)), transform=ax.transAxes)

        if legendOff is False:
            self.set_v22(galaxy)
            v22 = galaxy.v22
            plt.text(0.6,0.95,"V(2.2 Rd) %d km/s" % (v22),transform=ax.transAxes)

        if isinstance(galaxy, GalaxyParameters):
            if 'virial_velocity' in galaxy.names:
                Vvir = galaxy['virial_velocity']
                Mvir, Rvir = self.compute_MvirRvir(Vvir)
                logMvir=np.log10(Mvir)
                plt.axhline(Vvir, label='Vvir')
                #title += "Mvir %.2f " %(logMvir)
                if legendOff is False:
                    plt.text(0.6,0.9,"Vvir %d [km/s]" % (Vvir),transform=ax.transAxes)
                    plt.text(0.6,0.85,"Mvir %.2f [log Msun]" % (np.log10(Mvir)), transform=ax.transAxes)
                    plt.text(0.6,0.8, "Rvir %d [kpc] " % (Rvir), transform=ax.transAxes)

            elif 'maximum_velocity' in galaxy.names:
                vmax = galaxy['maximum_velocity']
                if legendOff is False:
                    plt.axhline(vmax, label='Vmax')
                    plt.text(0.6,0.9,"Vmax %d km/s" % (vmax),transform=ax.transAxes)


        #_xmin, _xmax = -0.1 * np.max(x)/3, 1.1*np.max(x)/3.
        #plt.xlim(_xmin, _xmax)
        if xmax is None:
            plt.xlim(-0.1*rhalf,5*rhalf)
        else:
            plt.xlim(-0.1*rhalf,xmax)
        if ymax is not None:
            plt.ylim(-10,ymax)


        if normalizeX == False:
            if self.redshift is not None and self.pixscale is not None:
                plt.xlabel('x (kpc)')
            else:
                plt.xlabel('x (pix)')
        else:
            plt.xlabel('r/Rhalf')
        #
        if legendOff is False:
            plt.legend(loc=2)
            plt.title(title)

        if filename is not None:
            plt.savefig(filename)
            plt.close()


    def predict_velocity(self, Xnew, chain, percentile=95):
        ci=[(100-percentile)/2.0, 50., 50+percentile/2.0 ]
        #mu = [np.median(self.set_velocity_profile(chain,x)) for x in Xnew]
        res = [np.nanpercentile(self.set_velocity_profile(chain,x),ci) for x in Xnew]
        res = np.array(res)

        #return median, low, upper range
        return res[:,1],res[:,0],res[:,2]

    def predict_SBprofile(self, Xnew, chain, percentile=95):
        ci=[(100-percentile)/2.0, 50., 50+percentile/2.0 ]
        #mu = [np.median(self.set_flux_profile(chain,x)) for x in Xnew]
        res = [np.nanpercentile(self.set_flux_profile(chain,x),ci) for x in Xnew]
        res = np.array(res)

        #return median, low, upper range
        return res[:,1],res[:,0],res[:,2]



import math


class NumericAssertions:
    """
    This class is a mixin following the UnitTest naming conventions.
    It is meant to be used along with unittest.TestCase like so :
    class MyTest(unittest.TestCase, NumericAssertions):
        ...
    It needs python >= 2.6
    """

    def assertIsNan(self, value, msg=None):
        """
        Fail if provided value is not nan
        """
        standardMsg = "%s is not nan" % str(value)
        try:
            if not math.isnan(value):
                self.fail(self._formatMessage(msg, standardMsg))
        except:
            self.fail(self._formatMessage(msg, standardMsg))

    def assertIsNotNan(self, value, msg=None):
        """
        Fail if provided value is nan
        """
        standardMsg = "Provided value is nan"
        try:
            if math.isnan(value):
                self.fail(self._formatMessage(msg, standardMsg))
        except:
            pass


# -*- coding: utf-8 -*-

class StringStdOut:
    """
    This is a dummy stdout to retrieve the output of asciitable.write().
    """
    def __init__(self):
        self.content = ''

    def write(self, content):
        self.content += content


# coding=utf-8
import numpy as np

try:
    from pyfftw.interfaces.numpy_fft import rfftn, irfftn, fftshift
except ImportError:
    from numpy.fft import rfftn, irfftn, fftshift


def convolve_nd_valid(cube, psf, compute_fourier=True):
    """
    using fft convolution of n-dim arrays M & N
    Returns a cube of dimensions (M-N+1)³
    """

    if compute_fourier:
        s1 = np.array(cube.shape)
        s2 = np.array(psf.shape)
    else:
        s1 = np.array(cube.shape)
        s2 = np.array(cube.shape) - np.array(psf.shape) + 1

    size_valid = s1 - s2 + 1

    axes = np.arange(s1.size)
    mdslice = list(axes * 0)  # empty list

    fsize = 2 ** np.ceil(np.log2(s1))
    padsize = fsize.astype('int32')  # must be integer
    mshape = fsize

    #output valid slices
    for i in axes:
        #mdslice[i]=slice(s1[i]-pad.shape[i]-size_valid[i],s1[i]-pad.shape[i])
        mdslice[i] = slice(s1[i] - padsize[i] - size_valid[i], s1[i] - padsize[i])

    if compute_fourier:
        fft_psf = rfftn(psf, s=mshape)
    else:
        fft_psf = psf

    res = irfftn(rfftn(cube, s=mshape) * fft_psf).real[mdslice]

    return res, fft_psf


def convolve_3d_same(cube, psf, compute_fourier=True):
    """
    Convolve a 3D cube with PSF & LSF.
    PSF can be the PSF data or its Fourier transform.
    if compute_fourier then compute the fft transform of the PSF.
    if False then assumes that the fft is given.

    This convolution has edge effects (and is slower when using numpy than pyfftw).

    cube: The cube we want to convolve
    psf: The Point Spread Function or its Fast Fourier Transform
    """

    # Pad to power of 2
    padded_cube, cube_slices = padding(cube, axes=[0, 1, 2])

    size = np.array(np.shape(padded_cube)[slice(0, 3)])

    if compute_fourier:
        padded_psf, psf_slices = padding(psf, axes=[0, 1, 2])
        fft_psf = rfftn(padded_psf, s=size, axes=[0, 1, 2])
    else:
        fft_psf = psf

    fft_img = rfftn(padded_cube, s=size, axes=[0, 1, 2])

    # Convolution
    fft_cube = np.real(fftshift(irfftn(fft_img * fft_psf, s=size, axes=[0, 1, 2]), axes=[0, 1, 2]))

    # Remove padding
    cube_conv = fft_cube[cube_slices]

    return cube_conv, fft_psf


def convolve_3d_xy(cube, psf, compute_fourier=True):
    """
    Convolve 3D cube along spatial directions only,
    using provided Point Spread Function.
    """

    # Compute needed padding
    cubep, boxcube = padding(cube, axes=[1, 2])

    size = np.array(np.shape(cubep)[slice(1, 3)])

    if compute_fourier:
        psfp, boxpsf = padding(psf, axes=[1, 2])
        fftpsf = np.fft.rfftn(psfp, s=size, axes=[1, 2])

    else:
        fftpsf = psf

    fftimg = np.fft.rfftn(cubep, s=size, axes=[1, 2])

    #Convolution
    fft = np.fft.fftshift(np.fft.irfftn(fftimg * fftpsf, s=size, axes=[1, 2]), axes=[1, 2]).real

    # Remove padding
    cube_conv = fft[boxcube]

    return cube_conv, fftpsf


def convolve_2d(image, psf, compute_fourier=True):
    """
    Compute fft of image And of PSF
    Accepts fftPSF already computed with do_PSF=False
    Reproduces Matt's convolution with PSF
    """

    # Compute needed padding
    cubep, boxcube = padding(image)

    if compute_fourier:
        psfp, boxpsf = padding(psf)
        fftpsf = np.fft.rfft2(psfp)
    else:
        fftpsf = psf

    fftimg = np.fft.rfft2(cubep)

    # Convolution
    fft = np.fft.fftshift(np.fft.irfft2(fftimg * fftpsf)).real

    # Remove padding
    cube_conv = fft[boxcube]

    return cube_conv, fftpsf


def convolve_1d(data, psf, compute_fourier=True, axis=0):
    """
    Convolve data with PSF only along one dimension specified by axis (default: 0)
    PSF can be the PSF data or its Fourier transform
    if compute_fourier then compute the fft transform of the PSF.
    if False then assumes that the fft is given.
    """

    axis = np.array([axis])

    # Compute needed padding
    cubep, boxcube = padding(data, axes=axis)

    # Get the size of the axis
    size = np.array(np.shape(cubep)[slice(axis, axis + 1)])

    if compute_fourier:
        psfp, boxpsf = padding(psf, axes=axis)
        fftpsf = np.fft.rfftn(psfp, s=size, axes=axis)
    else:
        fftpsf = psf

    fftimg = np.fft.rfftn(cubep, s=size, axes=axis)

    # Convolution
    fft = np.fft.fftshift(np.fft.irfftn(fftimg * fftpsf, s=size, axes=axis), axes=axis).real

    # Remove padding
    cube_conv = fft[boxcube]

    return cube_conv, fftpsf


def padding(cube, axes=None):
    """
    Computes padding needed for a cube to make sure it has
    a power of 2 shape along dimensions of passed axes (default [0,1])
    Returns padded cube and cube slices,
    which are the indices of the actual data in the padded cube.
    """

    if axes is None:
        axes = [0, 1]

    # Compute padding size for each axis
    old_shape = np.shape(cube)
    new_shape = np.array(old_shape)
    for axis in axes:
        zdim = cube.shape[axis]
        s = np.binary_repr(zdim - 1)
        s = s[:-1] + '0'
        new_shape[axis] = 2 ** len(s)

    cube_padded = np.zeros(new_shape)
    #cube_slices = np.empty(len(old_shape), slice).tolist()
    cube_slices = [None for i in range(len(old_shape))]

    for i, v in enumerate(old_shape):
        cube_slices[i] = slice(0, old_shape[i])

    for axis in axes:
        diff = new_shape[axis] - old_shape[axis]
        if (diff & 1):
            half = diff // 2 + 1
        else:
            half = diff // 2
        cube_slices[axis] = slice(half, old_shape[axis] + half)

    cube_slices = tuple(cube_slices)
    # Copy cube contents into padded cube
    cube_padded[cube_slices] = cube.copy()

    return cube_padded, cube_slices



# coding=utf-8


class AnsiColors:
    """
    A simple set of ANSI colors as escape sequences, stolen from Blender.
    """
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'


# -*- coding: utf-8 -*-
import numpy
from math import isnan
import logging, os
from astropy.table import Table

#local imports
from .galaxy_parameter import GalaxyParameter
from .string_stdout import StringStdOut
from .ansi_colors import AnsiColors

logging.basicConfig(level=logging.INFO)

class GalaxyParameters(numpy.ndarray):
    """
    A simple wrapper for an array of galaxy parameters.
    It is basically a flat numpy.ndarray of a bunch of floats with convenience
    attributes for explicit (and autocompleted!) access and mutation of
    parameters, and nice casting to string.

    .. warning::
        Undefined parameters will be NaN, not None.

    Example : ::

        from galpak import GalaxyParameters
        gp = GalaxyParameters(x=1.618)

        // Classic numpy.ndarray numeric access
        assert(gp[0] == 1.618)
        // Autocompleted and documented property access
        assert(gp.x == 1.618)
        // Convenience dictionary access,
        assert(gp['x'] == 1.618)
        // ... useful when you're iterating :
        for (name in GalaxyParameters.names):
            print gp[name]

    Parameters :
        - x (pix)
        - y (pix)
        - z (pix)
        - flux
        - radius aka. r½ (pix)
            half-light radius in pixel
        - inclination (deg)
        - pa (deg)
            position angle from y-axis, anti-clockwise.
        - turnover_radius aka. rv (pix)
            turnover radius for arctan, exp, velocity profile
            [is ignored for mass profile]
        - maximum_velocity aka. Vmax (km/s)
            de-projected V_max
            [forced to be positive, with 180deg added to PA]
        - velocity_dispersion aka. s0 (km/s)

    Note that all the parameters are lowercase, even if they're sometimes displayed mixed-case.

    Additional attributes :
        - stdev : Optional stdev margin (GalaxyParametersError).
                  GalPaK3D fills up this attribute.

    """

    # HOW TO CHANGE THE PARAMETERS
    #
    # Say we want to add a new galaxy parameter named `sugar` :
    # 1. Update the list `GalaxyParameters.names`, add 'sugar' at the end.
    # 2. Add the property
    #    sugar = GalaxyParameter(name='sugar', key=10, doc="Some more sugar", unit="sweetness", format="3.2f")
    #      - name (string) : the name of the variable, that MUST be the same as
    #                        the name in `GalaxyParameters.names`
    #      - key (int) : the position in the the ndarray of this parameter.
    #      - short (string) : a shorter name, for compact display.
    #      - doc (string) : some documentation that will appear to the enduser.
    #      - unit (string) : the unit, if any.
    #      - format (string) : the precision to use during string casting,
    #                             as used by `string.format()`.
    # 3. Update this class' __init__() method (signature and assignation)
    # 4. Update this class' __new__() method  (signature and call to __init__)

    names = [
        'x', 'y', 'z',
        'flux',
        'radius',
        'inclination',
        'pa',
        'turnover_radius',
        'maximum_velocity',
        'velocity_dispersion',
    ]

    x = GalaxyParameter(name='x', key=0, doc="X coordinates of the center.", unit="pixel")
    y = GalaxyParameter(name='y', key=1, doc="Y coordinates of the center.", unit="pixel")
    z = GalaxyParameter(name='z', key=2, doc="Z coordinates of the center.", unit="pixel", format='3.6f')
    flux = GalaxyParameter(name='flux', key=3, format='3.2e',
                           doc="Sum of the values of all pixels.", unit="same units as the input Cube")
    radius = GalaxyParameter(name='radius', short="rhalf", key=4, unit="pixel",
                             doc="Radius (half-light) of the galaxy.")
    inclination = pitch = GalaxyParameter(name='inclination', short="incl", key=5, unit="deg", format='4.1f',
                                          doc="Inclination, aka. pitch along observation axis.")
    pa = PA = roll = GalaxyParameter(name='pa', short="PA", key=6, unit="deg", format='5.1f',
                                     doc="Position Angle, anti-clockwise from Y (in degrees),"
                                         "aka. roll along observation axis.")
    turnover_radius = rv = GalaxyParameter(name='turnover_radius', short='rv', key=7, unit="pixel",
                                           doc="The turnover radius, inverse factor rv in arctan(r/rv).")
    maximum_velocity = max_vel = GalaxyParameter(name='maximum_velocity', short="Vmax", key=8, unit="km/s", format='6.2f',
                                                 doc="De-projected maximum velocity."
                                                     "Positive only, with 180deg added to PA.")
    velocity_dispersion = sigma0 = GalaxyParameter(name='velocity_dispersion', short="Vdisp", key=9, unit="km/s",
                                                   doc="Spatially constant disk dispersion,"
                                                       "added to kinematics dispersion.")

    # Developer notes :
    # There are many ways of doing the above :
    #   - decorators : @property, @prop.setter, etc.
    #   - __getattr__() and __setattr__()
    #   - __getattribute__() and __setattribute__()
    #   - explicit property()
    # We are using an extension of property() as it provides a cleaner way
    # to reuse getters and setters with multiple property names,
    # and provides autocompletion and customized docstrings.
    # Python dir() function is useful with __getattr__() and the likes, yet not
    # enough for proper documentation, and function wrappers are just horrible.
    # Another implementation of a similar feature :
    # https://github.com/numpy/numpy/blob/master/numpy/core/records.py

    ## CONSTRUCTORS
    stdev = None #standard deviation

    #confidence intervals
    ICpercentile = None
    lower = None
    upper = None

    """ Optional
    stdev margin (GalaxyParametersError)
    lower / upper bounds for a given percentile
    """

    def __init__(self,
                 x=None, y=None, z=None,
                 flux=None, radius=None,
                 inclination=None, pa=None,
                 turnover_radius=None,
                 maximum_velocity=None,
                 velocity_dispersion=None):


        self.x = x
        self.y = y
        self.z = z
        self.flux = flux
        self.radius = radius
        self.inclination = inclination
        self.pa = pa
        self.turnover_radius = turnover_radius
        self.maximum_velocity = maximum_velocity
        self.velocity_dispersion = velocity_dispersion

    def __new__(cls,
                x=None, y=None, z=None,
                flux=None, radius=None,
                inclination=None, pa=None,
                turnover_radius=None,
                maximum_velocity=None,
                velocity_dispersion=None):
        obj = super(GalaxyParameters, cls).__new__(cls, shape=len(GalaxyParameters.names), dtype=float)
        obj.__init__(
            x, y, z,
            flux, radius,
            inclination, pa,
            turnover_radius,
            maximum_velocity,
            velocity_dispersion
        )
        return obj

    ## NAMED CONSTRUCTORS (FACTORIES)

    @staticmethod
    def from_ndarray(a):
        """
        Factory to easily create a GalaxyParameters object from an ordered ndarray
        whose elements are in the order of `GalaxyParameters.names`.

        Use like this : ::

            from galpak import GalaxyParameters
            gp = GalaxyParameters.from_ndarray(my_ndarray_of_data)

        :rtype: GalaxyParameters
        """
        required_number_of_values = len(GalaxyParameters.names)
        if numpy.size(a, 0) < required_number_of_values:
            raise ValueError("Provide an array of at least %d floats (nan is allowed)" % required_number_of_values)
        return GalaxyParameters(*a)

    def save_to_file(self, filename, overwrite=False):
        """
        filename: file to save parameters
        overwrite: True/False to overwrite file
        """

        if not overwrite and os.path.isfile(filename):
            raise IOError("The file '%s' already exists. Specify overwrite=True to overwrite it.")
        else:
            t1 = self.as_table()
            t1.write(filename, format='ascii.fixed_width', overwrite=overwrite)

    def from_file(self, filename):
        """
        Reads the content from output.dat of GalaxyParameters
        """
        try:
            m = Table.read(filename, format='ascii.fixed_width')
            p = numpy.size(m)

            if p >= 2:
                self.stdev = GalaxyParameters()
            if p >= 3:
                self.lower = GalaxyParameters()
            if p >= 4:
                self.upper = GalaxyParameters()

            for k in self.names:
                self.__setitem__(k,m[k][0])
                if p >= 2:
                    self.stdev.__setitem__(k,m[k][1])
                if p >= 3:
                    self.lower.__setitem__(k,m[k][2])
                if p >= 4:
                    self.upper.__setitem__(k,m[k][3])

        except ImportError:
            raise ImportError("Need astropy module for this")

    def as_table(self):
        """
        generates a astropy Table
        :return:
        """
        t1 = Table(self, names=self.names)

        if self.stdev is not None:
            t1.add_row(self.stdev)
        if self.lower is not None and self.upper is not None:
            t1.add_row(self.lower)
            t1.add_row(self.upper)

        return t1

    def as_vector(self):
        """
        generates a vector
        :return: ndarray
        """
        vector = self.as_table()[0]
        return numpy.array(vector.as_void().tolist())

    #Hidden Extentions
    def set_v22(self, v22):
        #@fixme: can the be made into a new Parameter property?
        self._v22 = v22

    def get_v22(self):
        return self._v22

    def del_v22(self):
        del self._v22

    # class property([fget[, fset[, fdel[, doc]]]])
    v22 = property(get_v22, set_v22,  del_v22)

    #Hidden Extentions
    def set_dvdx(self, dvdx):
        #@fixme: can the be made into a new Parameter property?
        self._dvdx = dvdx

    def get_dvdx(self):
        return self._dvdx

    def del_dvdx(self):
        del self._dvdx

    # class property([fget[, fset[, fdel[, doc]]]])
    dvdx = property(get_dvdx, set_dvdx,  del_dvdx)


    ## MAGIC METHODS
    def __getitem__(self, key):
        """
        Support for dictionary-like accessors, like x = gp['x'].
        This has a (tiny) performance cost.
        """
        if isinstance(key, str):
            try:
                return self[self.names.index(key)]
            except ValueError:
                raise ValueError("Property '%s' is not available. Use one of %s" % (key, self.names))
        else:
            return super(GalaxyParameters, self).__getitem__(key)

    def __setitem__(self, key, value):
        """
        Support for dictionary-like mutators, like gp['x'] = x
        This has a (tiny) performance cost.
        """
        if isinstance(key, str):
            try:
                return super(GalaxyParameters, self).__setitem__(self.names.index(key), value)
            except ValueError:
                raise ValueError("Property '%s' is not available. Use one of %s" % (key, self.names))
        else:
            return super(GalaxyParameters, self).__setitem__(key, value)

    def __str__(self):
        return self.short_info()

    def __repr__(self):
        return self.long_info()

    def __unit__(self, key):
        """
        :return: shortname
        """
        if isinstance(key, str):
            try:
                return  vars(GalaxyParameters)[key].unit
            except:
                raise ValueError("Property '%s' is not available. Use one of %s" % (key, self.names))

    def unit_dict(self):
        """
         :return:  dict of short names for GalaxyParameters "
        """
        return dict([(n,vars(GalaxyParameters)[n].unit) for n in self.names])

    def __shortname__(self, key):
        """
        :return: shortname
        """
        if isinstance(key, str):
            try:
                return  vars(GalaxyParameters)[key].short
            except ValueError:
                raise ValueError("Property '%s' is not available. Use one of %s" % (key, self.names))

    def short_dict(self):
        """
         :return:  dict of short names for GalaxyParameters "
        """
        return dict([(n,vars(GalaxyParameters)[n].short) for n in self.names])

    ## STRING CASTING

    def short_info(self):
        """
        Casts to single-line string.
        This is called by ``print()``.

        Looks like this :
        x=0.00 y=1.00 z=2.000000 flux=3.00e+00 r½=4.00 incl=5.00 PA=6.00 rv=7.00 Vmax=8.00 s0=9.00

        :rtype: string
        """
        s = ''
        for name in GalaxyParameters.names:
            attribute = vars(GalaxyParameters)[name]
            s += ("{short}={value:%s} " % attribute.format).format(
                name=name,
                short=attribute.short,
                value=self.__getattribute__(name)
            )
        return s

    def colored_info(self, warn):
        """
        Casts to colored single-line string when GalaxyParameters `warn` is 1.
        The parameter `warn` is a GalaxyParameters object too, but it's values
        should be either 0 (no warning) or 1 (warning).

        Looks just like short info, but with additional ANSI characters for color
        where `warn` flag is 1.

        :rtype: string
        """
        s = ''
        for name in GalaxyParameters.names:
            attribute = vars(GalaxyParameters)[name]
            s += attribute.short + '='
            s += self._wrap_with_warn("{value:%s}" % attribute.format, warn.__getattribute__(name)).format(
                name=name,
                short=attribute.short,
                value=self.__getattribute__(name)
            )
            s += ' '
        return s

    def long_info(self, error_as_percentage=False):
        """
        Casts to multi-line string.
        This is called by ``repr()``.

        error_as_percentage: bool
            When true, will print the stdev margin as a percentage of the value.

        :rtype: string
        """
        if self.stdev is not None:
            s = "\n Galaxy Parameters : value (± stdev) [units]"
        else:
            s = "\n Galaxy Parameters : value [units]"

        if self.lower is not None and self.upper is not None:
            s += "\t [Confidence Interval]"
        s += '\n'

        for name in GalaxyParameters.names:
            prop = vars(GalaxyParameters)[name]
            value = self.__getattribute__(prop.name)
            # The name of the property, and its value
            s += ('    {name}: {value:%s}' % prop.format).format(
                name=prop.name,
                value=value
            )
            # The optional stdev margin
            if getattr(self, 'stdev', None) is not None:
                error_value = getattr(self.stdev, prop.name, None)
                if error_as_percentage:
                    error_value = 100 * error_value / value
                if not isnan(error_value):
                    s += (' ± {error_value:%s}' % prop.format).format(
                        error_value=error_value
                    )
                    if error_as_percentage:
                        s += '%'
            # And, finally, the optional unit
            if prop.unit is not None:
                s += " ({unit})".format(unit=prop.unit)
            if self.lower is not None and self.upper is not None:
                lower = getattr(self.lower, prop.name, None)
                upper = getattr(self.upper, prop.name, None)
                s += '\t CI '+str(self.ICpercentile)+'%:'
                s += (' [{lower:%s},{upper:%s}]  ' % (prop.format,prop.format)).format(
                    lower=lower,
                    upper=upper
                )

            s += '\n'

        return s

    def structured_info(self):
        """
        Casts to multiline parsable ``Table`` string.
        Use the ``ascii.fixed_width`` reader to easily retrieve the data.
        This method is used to store the parameters into the `<prefix>_galaxy_parameters.dat`.

        See the tutorials for an example of how to retrieve the data.
        """

        out = StringStdOut()

        t1 = self.as_table()
        t1.write(out,format='ascii.fixed_width')

        return str(out.content)

    @staticmethod
    def _wrap_with_warn(string, condition):
        """
        Will return string wrapped by WARN color escape sequence if condition is not zero.
        """
        if condition != 0 and not isnan(condition):
            return AnsiColors.WARNING + string + AnsiColors.ENDC
        else:
            return string


class GalaxyParametersError(GalaxyParameters):
    """
    Error margin for GalaxyParameters.
    We could just use another GalaxyParameters object instead of extending it,
    but this makes sure that stdev recursion never happens,
    and it also customizes the casting to string.
    """

    def get_error(self):
        return None

    def set_error(self, value):
        pass

    stdev = property(get_error, set_error, doc="""
    This property is useless.
    Error margins have no stdev margins.
    They don't, do they ?
    """)

    @staticmethod
    def from_ndarray(a):
        """
        Factory to easily create a GalaxyParametersError object from an ordered
        ndarray whose elements are in the order of `GalaxyParameters.names`.

        Use like this : ::

            gp = GalaxyParametersError.from_ndarray(my_ndarray_of_data)

        :rtype: GalaxyParametersError
        """
        required_number_of_values = len(GalaxyParameters.names)
        if numpy.size(a, 0) < required_number_of_values:
            raise ValueError("Provide an array of at least %d floats (nan is allowed)" % required_number_of_values)
        return GalaxyParametersError(*a)



import sys
import unittest
sys.path.append('../')

import numpy as np
from galpak import merge_where_nan, safe_exp

from tests.tools.numpy_ndarray_assertions import NumpyNdArrayAssertions


class MathUtilsTest(unittest.TestCase, NumpyNdArrayAssertions):

    longMessage = True

    def test_merge_where_nan(self):
        target = np.array([float('NaN'), 1.0])
        filler = np.array([2.0, 3.0])
        expected = np.array([2.0, 1.0])

        merge_where_nan(target, filler)

        self.assertArrayEqual(target, expected)

    def test_safe_exp(self):

        a = safe_exp(-1e100000000000000000)
        self.assertEqual(a, 0)

        a = safe_exp(-3950.)
        self.assertEqual(a, 0)

        a = safe_exp(1e1000000000000000000)
        self.assertEqual(a, float('inf'))

        a = safe_exp(3950.)
        self.assertEqual(a, float('inf'))





# -*- coding: utf-8 -*-

import numpy as np, time
from .galpak3d import GalPaK3D
from .model_sersic3d import ModelSersic as DefaultModel

# LOGGING CONFIGURATION
import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('GalPaK API')

def run(cube,
        variance=None,
        seeing=None, line=None, instrument=None,
        verbose=True,
        model=None,
        crval3=None, crpix3=None, cunit3=None, cdelt3=None, ctype3=None, cunit1=None,
        force_header_update=False,
        max_iterations=15000,
        method_chain='last',
        last_chain_fraction=60,
        percentile=95,
        mcmc_method='galpak',
        chi_stat = 'gaussian',
        min_boundaries=None,
        max_boundaries=None,
        known_parameters=None,
        initial_parameters=None,
        min_acceptance_rate=5.0,
        random_scale=None, **kwargs):
    """
    Creates a GalPaK3D instance and runs it on provided cube.
    Refer to documentations of GalPaK3D and GalPaK3D.run_mcmc().

    model: Model Class
        default: DefaultModel

    Use like this: ::

        import galpak
        g = galpak.run(cube='my.fits', instrument=myinstrument, model=mymodel)

        print g.galaxy
        g.save('my_run')
        # etc.
    """
    g = GalPaK3D(cube, variance=variance,
               instrument=instrument,  seeing=seeing,
               crval3=crval3, crpix3=crpix3, cunit3=cunit3, cdelt3=cdelt3, ctype3=ctype3, cunit1=cunit1,
               force_header_update=force_header_update)
    #default model
    if model is None:
        model = DefaultModel(line=line)
    g.run_mcmc(verbose=verbose,
               model=model,
               method_chain=method_chain,
               last_chain_fraction=last_chain_fraction,
               percentile=percentile,
               chi_stat=chi_stat,
               mcmc_method=mcmc_method,
               min_boundaries=min_boundaries,
               max_boundaries=max_boundaries,
               known_parameters=known_parameters,
               initial_parameters=initial_parameters,
               min_acceptance_rate=min_acceptance_rate,
               random_scale=random_scale,
               max_iterations=max_iterations, **kwargs)

    return g

def autorun(cube,
        variance=None,
        instrument=None,
        verbose=True,
        seeing=None,
        model=None,
        crval3=None, crpix3=None, cunit3=None, cdelt3=None, ctype3=None, cunit1=None,
        force_header_update=False,
        max_iterations=5000,
        method_chain='last',
        last_chain_fraction=60,
        percentile=95,
        chi_stat = 'gaussian',
        mcmc_method='galpak',
        min_boundaries=None,
        max_boundaries=None,
        known_parameters=None,
        initial_parameters=None,
        random_scale=1,
        MaxIter=25,
        target_acceptance = 40,
        target_tolerance = 5, **kwargs):
    """
    Creates a GalPaK3D instance and runs it on provided cube.
    Refer to documentations of GalPaK3D and GalPaK3D.run_mcmc().

    model: Model Class
        default: ModelSersic

    MaxIter: int
        default: 25
        maximum number of iterations for searching rscale
    target_acceptance : float
        default: 40
        target acceptance rate

    Use like this: ::

        import galpak
        g = galpak.autorun(cube='my.fits', instrument=myinstrument, model=mymodel)

        print g.galaxy
        g.save('my_run')
        # etc.
    """
    g = GalPaK3D(cube, variance=variance,
                instrument=instrument, seeing=seeing,
                quiet=True,
                crval3=crval3, crpix3=crpix3, cunit3=cunit3, cdelt3=cdelt3, ctype3=ctype3, cunit1=cunit1,
                force_header_update=force_header_update)
    #default model
    if model is None:
        model = DefaultModel()

    #first quick run
    logger.info("First quick run")
    g.run_mcmc(verbose=None,
               model=model,
               method_chain=method_chain,
               mcmc_method='galpak',
               last_chain_fraction=last_chain_fraction,
               percentile=percentile,
               chi_stat=chi_stat,
               min_boundaries=min_boundaries,
               max_boundaries=max_boundaries,
               known_parameters=known_parameters,
               initial_parameters=initial_parameters,
               random_scale=random_scale,
               max_iterations=250)
    #iterate to calibrate random_scale
    logger.info("Initial Acceptance %.2f" % (g.acceptance_rate))

    if g.acceptance_rate>80:
        g.random_scale = 15

    i=0
    f=1
    new_rscale=1.0
    if isinstance(g.random_scale,np.ndarray):
        raise NotImplementedError("Random Scale must be float for autorun")

    while (i<5) or ((np.abs(g.acceptance_rate-target_acceptance)>target_tolerance) and (i<MaxIter)):

        new_rscale= g.random_scale+f*np.log10(g.acceptance_rate/target_acceptance)

        if i>10 and new_rscale>1:
            f=2 #adapt faster
        if new_rscale<0:
            f=0.25
            new_rscale = g.random_scale + f * np.log10(g.acceptance_rate / target_acceptance)
        #if new_rscale<0.5:
        #    f / 2
        if new_rscale>15:
            f=5

        #@fixme check if random_scale becomes <0
        if new_rscale<0:
            new_rscale=0.5 #minimum
            i = MaxIter
        g.run_mcmc(verbose=None,
               model=model,
               method_chain=method_chain,
               mcmc_method='galpak',
               last_chain_fraction=last_chain_fraction,
               percentile=percentile,
               chi_stat=chi_stat,
               min_boundaries=min_boundaries,
               max_boundaries=max_boundaries,
               known_parameters=known_parameters,
               initial_parameters=g.galaxy,
               random_scale=new_rscale,
               max_iterations=150);
        logger.info("%d Rscale %.2f Acceptance %.2f%% Target: %d%%" \
                             % (i,g.random_scale, g.acceptance_rate, target_acceptance))
        logger.info(" Parameters : " + g.galaxy.short_info())
        i+=1
    #save rscale
    tuned_rscale = g.random_scale
    logger.info("Using best Rscale %.2f" % (tuned_rscale))    #reusing g.galaxy
    time.sleep(5)#wait 5sec
    #then run everything
    if i<MaxIter:
        g.run_mcmc(verbose=verbose,
               model=model,
               method_chain=method_chain,
               mcmc_method=mcmc_method,
               last_chain_fraction=last_chain_fraction,
               percentile=percentile,
               chi_stat=chi_stat,
               min_boundaries=min_boundaries,
               max_boundaries=max_boundaries,
               known_parameters=known_parameters,
               initial_parameters=g.galaxy,
               random_scale=tuned_rscale,
               max_iterations=max_iterations, **kwargs)
        logger.info(" Parameters : " + g.galaxy.short_info())
        return g
    else:
        logger.error("AutoRandomScale Failed")

        return None



import benchmark
import numpy as np

#Benchmark Report
#================
#
#BenchmarkPowerOfThree
#---------------------
#
#                   name | rank |  runs |      mean |        sd | timesBaseline
#------------------------|------|-------|-----------|-----------|--------------
#explicit multiplication |    1 | 1e+04 | 0.0002208 | 4.511e-05 |           1.0
#            numpy power |    2 | 1e+04 |  0.003243 | 0.0002166 | 14.6868697001
#       double asterisks |    3 | 1e+04 |  0.003244 | 0.0002177 |  14.692235388
#
#Each of the above 30000 runs were run in random, non-consecutive order by
#`benchmark` v0.1.5 (http://jspi.es/benchmark) with Python 2.7.5+
#Linux-3.11.0-12-generic-x86_64 on 2014-02-13 14:45:30.


class BenchmarkPowerOfThree(benchmark.Benchmark):

    each = 10000  # configure number of runs

    def setUp(self):
        # Can also specify tearDown, eachSetUp, and eachTearDown
        self.a = np.random.rand(32, 32, 32)

    def test_explicit_multiplication(self):
        b = self.a * self.a * self.a

    def test_double_asterisks(self):
        b = self.a ** 3

    def test_numpy_power(self):
        b = np.power(self.a, 3)


if __name__ == '__main__':
    benchmark.main(format="markdown", numberFormat="%.4g")




# coding=utf-8

## GENERAL PACKAGES
import os
import unittest

import numpy as np

import logging
logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger('GalPaK Q-test')

from astropy.io import fits

import galpak
from galpak import GalPaK3D
from galpak import GalaxyParameters
from galpak import MUSEWFM
from galpak import HyperspectralCube as Cube

root_folder = os.path.dirname(os.path.abspath(galpak.__file__))
fits_folder = os.path.join(root_folder, '../../data/input/ref')

from tests.tools.numpy_ndarray_assertions import NumpyNdArrayAssertions


class GalPaK3DQuantuumTest(unittest.TestCase, NumpyNdArrayAssertions):
    """
    These are miscellaneous tests that may or may not pass. (hence the name)
    They're kept because they are useful, but they're not part of the testing suite.
    You are NOT supposed to run them, but you may, at your expense and risk.
    WARNING : some of these tests take a very long time (they run many deconvolutions) !
    """

    longMessage = True

    fits_muse_filename = os.path.join(fits_folder, 'GalPaK_cube_1101_from_v191_noised.fits')

    def assertArrayEqualishNico(self, expected, actual, expected_similarity=99.):
        """
        expected_similarity (in %)
        """
        # Avoid rounding errors and infinity errors
        good = lambda x: np.where(np.abs(x) > 1e-32, np.float32(x), 1e-32)
        expected_good = good(expected)
        actual_good = good(actual)
        differences = np.abs((expected_good - actual_good) / actual_good)
        #differences = good(differences)
        differences_not_nan = differences[~np.isnan(differences)]
        actual_precision = 100. * (1 - np.mean(differences_not_nan))
        self.assertTrue(actual_precision >= expected_similarity, "Expected %f%%, got %f%%" % (expected_similarity, actual_precision))
        return actual_precision


    def test_psf3d(self):
        expected_similarity = 99.999  # in %

        psf_cube_file = os.path.join(fits_folder, 'psf3d_current_pair_Seeing08.fits')
        psf_expected = fits.open(psf_cube_file)[0].data

        instr = MUSEWFM(psf_fwhm=0.8)

        glpk3d = GalPaK3D(Cube(data=np.ones((20,30,30))), instrument=instr)

        image = instr.psf.as_image(glpk3d.cube)
        lsf1d = instr.lsf.as_vector(glpk3d.cube)
        #self.psf.as_image(cube), self.lsf.as_vector(cube), cube
        psf3d_galpak = instr.extrude_psf(image, lsf1d)

        print("PSF3D precision =", self.assertArrayEqualish(psf_expected, psf3d_galpak, expected_similarity))

    def test_galfitness(self):
        """
        Compare GalPaK3D results with galfit results
        This is a debugging test
        """
        expected_similarity = 98.  # in %
        clean_cube_file = os.path.join(fits_folder, 'cubeclean_current_pair-2.fits')
        conv_cube_file = os.path.join(fits_folder, 'cubetest_current_pair_Seeing0.8.fits')

        # P=[15,15,11,3,55,45,1,250,100,0]
        params = GalaxyParameters.from_ndarray([15, 15, 11, 100, 3 * 1.6721, 55, 45, 1, 250, 0])

        glpk3d = GalPaK3D(Cube(data=np.ones((20,30,30))), instrument=MUSEWFM(psf_fwhm=0.8))
        glpk3d.model = galpak.DiskModel()

        cube_clean = glpk3d.create_clean_cube(params, (20, 30, 30))
        cube_clean_expected = fits.open(clean_cube_file)[0]

        print("Clean precision =", self.assertArrayEqualishNico(cube_clean_expected.data, cube_clean.data, expected_similarity))

        cube_conv_expected = fits.open(conv_cube_file)[0]
        cube_conv = glpk3d.create_convolved_cube(params, (20, 30, 30))

        print("Conv precision =", self.assertArrayEqualish(cube_conv_expected.data, cube_conv.data, expected_similarity))
        #print "Conv precision =", self.assertArrayEqualishNico(cube_conv_expected.data, cube_conv.data, expected_similarity)





